package Weigel::Search;

use 5.008001;
use strict;
use warnings;

use Weigel::Search::Backends::Vmatch;
use Weigel::Search::Backends::Agrep;
use Weigel::Search::Backends::Hypa;

use base 'Bio::Root::Root';

our $VERSION = '0.03';

use Class::MethodMaker [
   new      => 'new2',
   scalar   => [qw / backend/],
];

=head1 NAME

Weigel::Search - Perl extension for searching in Fasta files

=head1 SYNOPSIS

  use Weigel::Search;
  
  use Bio::Root::Exception;
  use Error qw(:try);
  
  my $backend = "Hypa";
  
  my $search_obj = Weigel::Search->new($backend);	
  
  # $sbe is now a reference to the backend
  # perldoc Weigel::Search::Backend::BackendI	
  # perldoc Weigel::Search::Backend::Vmatch
  my $sbe = $search_obj->backend;
  
  $sbe->settings->tmppath('tmp');
  $sbe->settings->datapath('data');
  
  mkdir($sbe->settings->tmppath);	
  mkdir($sbe->settings->datapath);	
  
  # generate a suffix array. you have to do this only once.
  #$sbe->generate_database_out_of_fastafile('../t/Test.fasta', 'Description for the test Fastafile');
  
  $sbe->settings->database('Test.fasta');
  
  my $seq = 'UGAACAGAAAG';
  
  # search for the reverse complement and allow 2 mismatches
  $sbe->settings->query($seq);
  $sbe->settings->reverse_complement(1);
  $sbe->settings->mismatches(2);
  
  # now try to search. we use bioperls Exceptions
  try {
     $sbe->search();
  } catch Bio::Root::SystemException with {
     my $E = shift;
     print STDERR 'Backend call failed: ' . 	
     $E->{'-text'} . ' (' .  $E->{'-line'} . ")\n";
     exit(1);	
  } otherwise {        
     my $E = shift;
     print STDERR "An unexpected exception occurred: \n$E";
     exit(1);	
  };
  
  print scalar(@{$sbe->results}) . " search results\n";
  
  # output some informations! 
  # perldoc Weigel::Search::Container::SearchResult
  
  foreach my $res (@{$sbe->results}) {
     print $res->sequence->id . "\n";
     print $res->mark_subject_uppercase() . "\n";
     print $res->alignment_string() . "\n\n";
  }

=head1 DESCRIPTION

Weigel-Search is a collection of Perl modules for searching in 
FASTA-files. It is programmed in a modular way. There are different 
backends available. You can filter search results.

There are other modules available for evaluating search results.
Weigel::RNA for example can filter putative MicroRNA targets. 

=head1 METHODS

=over 2 

=item new($backend)

This function constructs a C<Weigel::Search> object. Available backends
are Vmatch, Agrep and Hypa. Vmatch is default.

=cut

sub new {
   my $self = shift->new2;
   my $backendname  = shift;
   $backendname = 'Vmatch' unless defined $backendname;

   unless ($backendname eq 'Agrep' || $backendname eq 'Vmatch' || $backendname eq 'Hypa') {
      $self->throw(-class => 'Bio::Root::BadParameter',
                   -text  => 'Unknown backend.',
                   -value => $backendname . ' not supported.');
   }
   my $backendpackage = "Weigel::Search::Backends::$backendname";
   
   $self->backend( $backendpackage->new() );
   $self;
}

=item backend()

Get/set the backend. This is a objects that uses L<Weigel::Search::Backend::BackendI> 
as base class. See L<Weigel::Search::Backends::BackendI>, L<Weigel::Search::Backends::Vmatch>,
L<Weigel::Search::Backends::Agrep> and L<Weigel::Search::Backends::Hypa>

=head1 SEE ALSO

L<Weigel::Search::Backends::BackendI>
L<Weigel::Search::Backends::Vmatch>
L<Weigel::Search::Backends::Agrep>
L<Weigel::Search::Backends::Hypa>

L<http://weigelworld.org/resources/software/perl/>

=head1 AUTHOR

Markus Riester, E<lt>markus.riester@cgw.tuebingen.mpg.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005 by markus

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.4 or,
at your option, any later version of Perl 5 you may have available.


=cut

1;

# vim: ft=perl sw=3 ts=3 expandtab
