package Weigel::Search::Backends::Agrep;

use strict;
use warnings;

use Weigel::Search::Container::SearchSettings;
use Weigel::Search::Container::SearchResult;
use Weigel::Search::Backends::BackendI;

use base 'Weigel::Search::Backends::BackendI';

use Bio::Seq;
use Bio::SeqIO;
use Bio::Index::Fasta;
##use String::Approx 'aindex';

use File::Basename;
use IO::String;
use Cwd;

our $VERSION = '0.2';

=head1 NAME

Weigel::Search::Backends::Agrep.pm - Agrep Backend  

=head1 SYNOPSIS

  use Weigel::Search::Backends::Agrep;
  
  use Error qw(:try);
  
  # configure our search backend, in this case Agrep
  my $sbe = Weigel::Search::Backends::Agrep->new();
  
  $sbe->settings->execpath('/usr/bin');
  $sbe->settings->tmppath('tmp');
  $sbe->settings->datapath('data_agrep');
  
  # generate a database. you have to do this only once. 
  $sbe->generate_database_out_of_fastafile('ATH1.cdna', 'AGI Transcripts (- introns, + UTRs)');
  
  my %local_dbs_description = $sbe->get_databases();
  my @local_dbs = sort keys %local_dbs_description;
  
  # take first available database in our test
  $sbe->settings->database($local_dbs[0]);
  
  my $seq = 'UGAACAGAAAGCUCAUGAGCC'; 
  
  # search for the reverse complement and allow 4 mismatches
  $sbe->settings->query($seq);
  $sbe->settings->reverse_complement(1);
  $sbe->settings->mismatches(4);
  
  # things should be fast? then turn alignment calculation off
  $sbe->settings->no_alignments(1);
  
  # now try to search
  try {
     $sbe->search();
  } catch Error::Simple with {
     my $E = shift;
     print  $E->{'-text'} . ' (' .  $E->{'-line'} . ")";
  };
  
  my @results = @{$sbe->results};
  print scalar(@results) ." search results\n\n";
  # output the searchresults with nice alignments
  foreach my $res (@results) {
     print $res->sequence->id . "\n";
     print $res->mark_subject_uppercase() . "\n";
     print $res->alignment_string() . "\n\n";
  }
  
=head1 DESCRIPTION

B<Weigel::Search::Backends::Agrep> searches for a query with agrep. 

=head1 METHODS

See L<Weigel::Search::Backends::BackendI> for other methods. 

=over 2

=item Weigel::Search::Backends::Agrep-E<gt>new()

This function constructs a Vmatch Backend object

   my $sbe = Weigel::Search::Backends::Agrep-->new();

=cut

sub new {
   my $self = shift;
   $self = $self->SUPER::new;
   my %all_features = $self->features;
   delete $all_features{NATIVE_ALIGNMENTS};
   delete $all_features{GUMISMATCHES};
   delete $all_features{FILTERS};
   delete $all_features{EVALUE};
   delete $all_features{PERCENT_IDENTITY};
   delete $all_features{ONLINE};
   $self->features(%all_features);
   $self;
}

=item $sbe-E<gt>search()

This function searches for the query specified in the L<Weigel::Search::Container::SearchSettings> object C<$sbe-E<gt>settings>
 
=cut

sub search {
   my $self = shift;
   my $s    = $self->settings;

   $self->warn("GU mismatch options not available in this backend")
     if $s->gumismatches_isset && $s->gumismatches != 1;
   $self->warn("Filter options not available in this backend")
     if $s->filter_isset;
   
   $self->_prepare_query();
   $self->_check_search_settings();

   # now generate the command string
   my $mm = 0;
   $mm = $s->mismatches if $s->mismatches > 0;

   # make insertions and deletions to expensive
   my $fuzzy = "-$mm -D" . ( $mm + 1 ) . " -I" . ( $mm + 1 );

   if ( $s->editdistance_isset ) {
      $fuzzy = "-" . $s->editdistance;
      $self->warn("Ignoring mismatch settings because edit distance is set.")
        if $s->mismatches > 0;
   }

   $s->querylength( length( $s->query ) ) unless $s->querylength_isset;
   my $length = "-l " . $s->querylength;

   my $command =
       $self->_cat_path_filename( $s->execpath, 'agrep' )
     . ' -i -n -t '
     . $fuzzy . ' '
     . $s->query . ' '
     . $self->_cat_path_filename( $s->datapath, $s->database . '.dat' );

   #  print $command ."\n";
   my $output = `$command`;
   $self->throw( -class => 'Bio::Root::SystemException', 
                 -text  => "Agrep error: Query not valid.\n\nCommand was:\n\t$command"
     )
     if ($?);
   $self->_output_parser($output);
}

=item $sbe->get_databases()

Returns a hash with all available databases. The keys are the filenames,
the values are descriptions (or the filename if no description is available).

Descriptions can be set in info files. For example, if you indexed file ATH1.cdna, vmatch constructs a lot of ATH1.cdna.* files. Now simply create a file 
ATH1.cdna.nfo and write a description in that file. The function generate_database_out_of_fastafile will
create this file for you if you add a description as second argument.
 
=cut

sub get_databases {
   my $self = shift;
   return $self->_get_databases('.map');
}

=item $sbe->generate_database_out_of_fastafile('/tmp/ATH1.cdna', 'AGI Transcripts (- introns, + UTRs)')

Copies the specified file in the datapath directory ($sbe->settings->datapath)
and generate some files for the agrep program. You can get the databases with $sbe->get_databases().
The second argument is a little description for the database.

=cut

sub generate_database_out_of_fastafile {
   my $self        = shift;
   my $file        = shift;
   my $description = shift;
   my ( $filename, $oldpath ) = fileparse($file);

   my $olddir = getcwd();
   
   $self->_copy_fasta_file_and_create_nfo($file, $filename, $description);

   open( DATFILE, "> $filename.dat" )
     or $self->throw(-class => 'Bio::Root::FileOpenException',
                     -text  => "Can't open $filename.dat for writing",
                     -value => $!);

   open( MAPFILE, "> $filename.map" )
     or $self->throw(-class => 'Bio::Root::FileOpenException',
                     -text  => "Can't open $filename.map for writing",
                     -value => $!);

   my $in = Bio::SeqIO->new( -file => $filename, -format => 'Fasta' );
   my $i = 0;
   while ( my $seq = $in->next_seq() ) {
      print DATFILE $seq->seq . "\n";
      print MAPFILE $i . "\t" . $seq->id . "\n";
      $i++;
   }
   close(DATFILE);
   close(MAPFILE);

   # create database from directory of fasta files
   my $idx = Bio::Index::Fasta->new(
      -filename   => $filename . '.idx',
      -write_flag => 1
   );
   $idx->make_index( ($filename) );

   chdir($olddir);
}

sub _output_parser {
   my $self    = shift;
   my @lines   = split( "\n", shift );
   my $s       = $self->settings;
   my @results = ();

   my $mapfile   = $s->datapath . '/' . $s->database . '.map';
   my $indexfile = $s->datapath . '/' . $s->database . '.idx';

   my $idx = Bio::Index::Fasta->new($indexfile);

   my %mapping = ();

   open( MAPFILE, $mapfile )
     or $self->throw(-class => 'Bio::Root::FileOpenException',
                     -text  => "Can't open $mapfile for reading",
                     -value => $!);

   while (<MAPFILE>) {
      chomp;
      my ( $pos, $id ) = split("\t");
      $mapping{$pos} = $id;
   }
   close(MAPFILE);

   foreach (@lines) {
      my ( $pos, $sequence ) = split(":");
      my $id = $mapping{$pos};

      my $seq_subject = $idx->fetch($id);

      my $seq_query = Bio::Seq->new(
         -display_id => "Query",
         -seq        => $s->query
      );

      # my $seq_subject = Bio::Seq->new(
      #    -display_id => "Subject",
      #    -seq        => $query
      # );

      my $alignment = undef;
      $alignment = $self->_get_alignment( $seq_query, $seq_subject )
        unless $s->no_alignments;

      my $result =
        Weigel::Search::Container::SearchResult->new( $seq_subject,
         $s->upstream, $s->upstream + $s->querylength,
         $alignment, $seq_subject->id, "" );
      push( @results, $result );
      $self->results(@results);
   }
}

#TODO not yet finished

=item $sbe-E<gt>get_sequences($sequence_ids) 

This function returns all sequences with the ids in the specified array reference as
a L<Bio::SeqIO> object. 


   my $seqio = $sbe->get_sequences([$id]);
   my $string;  my $stringio = IO::String->new($string);
   my $out = Bio::SeqIO->new('-fh' => $stringio,
                             '-format' => 'fasta');

   while ( my $seq = $seqio->next_seq() ) {
      # write the sequences in a string
      $out->write_seq($seq);
   }
   print $string;
 

=cut

sub get_sequences {
   my $self  = shift;
   my $seqid = shift;
   my $indexfile = $self->settings->datapath . '/' . $self->settings->database . '.idx';
   my $idx = Bio::Index::Fasta->new($indexfile);
   my $out = Bio::SeqIO->new( -format => 'Fasta' );
   foreach my $seqid ( @{$seqid} ) {
      $out->write_seq($idx->fetch($seqid));
   }
   return $out;
}

=back


=head1 SEE ALSO

L<Weigel::Search::Backends::BackendI>
L<Weigel::Search::Container::SearchSettings>
L<Bio::SeqIO>
L<Bio::Index::Fasta>

=head1 AUTHOR

Markus Riester, <markus.riester@tuebingen.mpg.de@>

=cut

1;

# vim: ft=perl sw=3 ts=3 expandtab
