package Weigel::Search::Backends::BackendI;

use strict;
use warnings;

use Weigel::Search::Container::SearchSettings;
use Weigel::Search::Container::SearchResult;

use Bio::Seq;
use Bio::SeqIO;
use Bio::AlignIO;
use Bio::Factory::EMBOSS;

use base 'Bio::Root::Root';

use File::Spec;
use File::Copy;

our $VERSION = '0.3';

use Class::MethodMaker [
   new      => 'new2',
   scalar   => [qw / settings /],
   array    => [qw / results /],
   hash     => [qw / features /],
   abstract => [qw / search get_sequences get_databases generate_database_out_of_fastafile /],
];

=head1 NAME

Backends::Backend.pm - superclass for all backends  

=head1 SYNOPSIS

See the backend modules for example code.

=head1 DESCRIPTION

This is the superclass for all backends.

=head1 METHODS

=over 2 

=item new()

This function constructs a Backend object

=cut

sub new {
   my $self = shift->new2;

   # initialize standard settings
   my $settings = Weigel::Search::Container::SearchSettings->new();

   $self->settings($settings);

   # assume backend binary is in path
   $self->settings->execpath('');

   my %all_features = ( MISMATCHES => 1, GUMISMATCHES => 1, EDITDISTANCE => 1,
                        INSERTIONS => 1, DELETIONS => 1, FILTERS => 1,
                        NATIVE_ALIGNMENTS => 1, EVALUE => 1, PERCENT_IDENTITY => 1,
                        ONLINE => 1,
                      );
   $self->features(%all_features);
   $self;
}

=item _get_databases($suffix)

This function searches the data directory for files ending with $suffix
and returns this list of files in an array

Substitutes $suffix with .nfo from all found files and searches for an
info file with that name. The content of that file will be used as description.
When no file is found, the description will be the filename without the suffix:

=over 4
%dbs = _get_databases('.al1');  # finds file ATH1.cdna.al1, searches for ATH1.cdna.nfo
print $dbs{'ATH1.cdna'};      # prints content of ATH1.cdna.nfo or 'ATH1.cdna'
=back

=cut

sub _get_databases {
   my $self   = shift;
   my $suffix = shift;
   my %result = ();
   opendir( DIR, $self->settings->datapath );
   my @files = grep { /${suffix}$/ } readdir(DIR);
   closedir DIR;
   foreach my $file (@files) {
      my $prefix = $file;
      $prefix =~ s/$suffix//;
      $result{$prefix} = $prefix;
      $file =~ s/$suffix/\.nfo/;
      $file = $self->settings->datapath . '/' . $file;
      open( FILE, $file ) or next;
      my $desc = "";
      while (<FILE>) {
         chomp();
         $desc .= $_;
      }
      $result{$prefix} = $desc;
      close(FILE);
   }
   return %result;
}

sub _filter_results {
   my $self = shift;
   if ($self->settings->filter_isset) {
      my @filtered = ();
      my $filter = $self->settings->filter;
      foreach my $res (@{$self->results}) {
         $filter->query( $self->settings->query );
         $filter->target( $res->sequence->subseq($res->begin+1, $res->end) );
         unless ( $filter->filter ) {
            if ( $filter->delete ) {
               next;
            }
         }
         $res->remark($filter->message);
         push(@filtered, $res);
      }
      $self->results(@filtered);
   }
}

sub _get_alignment {
   my $self    = shift;
   my $seq_a   = shift;
   my $seq_b   = shift;
   my $factory = Bio::Factory::EMBOSS->new();
   my $prog    = $factory->program('needle');
   my $outfile = $self->settings->tmppath . "/" . $seq_a->id . ".out";
   my @seqs    = qw($seq_b);
   my $gapopen = '5.0';
   $gapopen = '50.0' unless $self->settings->editdistance_isset;

   $prog->run(
      {
         -asequence => $seq_a,
         -bsequence => $seq_b,
         -gapopen   => $gapopen,
         -gapextend => '5.0',
         -outfile   => $outfile
      }
   );

   if ( -e $outfile ) {
      my $alignio_fmt = "emboss";
      my $align_io    = new Bio::AlignIO(
         -format => $alignio_fmt,
         -file   => $outfile
      );
      unlink($outfile);
      return $align_io->next_aln();
   }

}

sub _cat_path_filename {
   my $self = shift;
   my $path = shift;
   my $filename = shift;
   return $filename if $path eq '';
   return File::Spec->catfile($path , $filename);
}

# some basic error checking. important security checks, because
# we use system(). so we should check, if we get what we assume.
sub _check_search_settings {
   my $self = shift;
   $self->settings->upstream(0) unless $self->settings->upstream_isset;
   $self->settings->downstream(0) unless $self->settings->downstream_isset;
   
   $self->throw( -class => 'Bio::Root::BadParameter',
                 -text  => 'Upstream not an integer',
                 -value => 'upstream' ) unless $self->settings->upstream =~ /^\d+$/;
   $self->throw( -class => 'Bio::Root::BadParameter',
                 -text  => 'Downstream not an integer',
                 -value => 'downstream' ) unless $self->settings->downstream =~ /^\d+$/;

   if ($self->settings->mismatches_isset) {
      $self->throw( -class => 'Bio::Root::BadParameter',
                    -text  => 'Mismatches not an integer',
                    -value => 'mismatches' ) unless $self->settings->mismatches =~ /^\d+$/;
   }
   if ($self->settings->insertions_isset) {
      $self->throw( -class => 'Bio::Root::BadParameter',
                    -text  => 'Insertions not an integer',
                    -value => 'insertions' ) unless $self->settings->insertions =~ /^\d+$/;
   }
   if ($self->settings->deletions_isset) {
      $self->throw( -class => 'Bio::Root::BadParameter',
                    -text  => 'Deletions not an integer',
                    -value => 'deletions' ) unless $self->settings->deletions =~ /^\d+$/;
   }
   # _prepare_query should check query
}

# changes directory! so please save the oldpath
sub _copy_fasta_file_and_create_nfo {
   my $self = shift;
   my $file = shift;
   my $filename = shift;
   my $description = shift;

   my $newfile =
     $self->_cat_path_filename( $self->settings->datapath, $filename );

   copy( $file, $newfile )
     or $self->throw( -class => 'Bio::Root::IOException',
                      -text  => "Can't copy $file to $newfile",
                      -value =>  $!
     );

   chdir( $self->settings->datapath )
     or $self->throw( -class => 'Bio::Root::IOException',
                      -text  => "Can't change to directory " . $self->settings->datapath,
                      -value => $! );

   if ( defined($description) ) {
      open( NFOFILE, ' > ' . $filename . '.nfo' )
        or $self->throw(-class => 'Bio::Root::FileOpenException',
                        -text  => "Can't open $filename.nfo for writing"
                        -value => $!);
      print NFOFILE $description;
      close(NFOFILE);
   }
}

sub _prepare_query {
   my $self = shift;
   my $query = $self->settings->query;
   $query =~ tr/uU/tT/;
   my $seq = Bio::Seq->new( -seq => $query);
   $self->settings->reverse_complement(0) unless $self->settings->reverse_complement_isset;
   $query = $seq->revcom->seq if $self->settings->reverse_complement;
   $self->settings->query(uc($query));
}

=item settings()

Get/set the settings. This is a L<Weigel::Search::Container::SearchSettings> object

=item results()

Get/set the results. This is an array of L<Weigel::Search::Container::SearchResults> objects.

=item warnings()

Get/set the warnings. This is an array containing warning strings.

=item features()

Get/set available features. This is an hash. Valid features are
MISMATCHES, GUMISMATCHES, EDITDISTANCE, INSERTIONS, DELETIONS, 
FILTERS, NATIVE_ALIGNMENTS, EVALUE and PERCENT_IDENTITY.

  if (defined($sbe->features->{GUMISMATCHES})) {
          # $sbe->settings->gumismatches(0);
          $sbe->settings->gumismatches(0.5);
  } else {
        print "\nBackend does not support wobble pairs\n";
  }

   
=head1 FILES

Requires EMBOSS and Bio::Factory::EMBOSS for the Needleman-Wunsch local 
alignment implementation from EMBOSS. The private function 
_get_alignment($seq_a, $seq_b) can than calculate an alignment for 
backends that do not generate a alignment (like HYPA, agrep).

=head1 SEE ALSO

L<Weigel::Search::Container::SearchSettings> 
L<Weigel::Search::Container::SearchResults> 

=head1 AUTHOR

Markus Riester, <markus.riester@tuebingen.mpg.de@>

=cut

1;

# vim: ft=perl sw=3 ts=3 expandtab
