package Weigel::Search::Backends::Hypa;

use strict;
use warnings;

use Weigel::Search::Container::SearchSettings;
use Weigel::Search::Container::SearchResult;
use Weigel::Search::Backends::BackendI;

use base 'Weigel::Search::Backends::BackendI';

use Bio::Seq;
use Bio::SeqIO;

use File::Basename;
use IO::String;
use Cwd;

our $VERSION = '0.1';

=head1 NAME

Backends::Hypa.pm - Hypa Backend  

=head1 SYNOPSIS

  use Weigel::Search::Backends::Hypa;
 
  use Bio::Root::Exception;
  use Error qw(:try);
 
  # construct the Hypa backend	
  my $sbe = Weigel::Search::Backends::Hypa->new();
 
  $sbe->settings->tmppath('tmp');
  $sbe->settings->datapath('data');
  
  # generate a Hypa suffix array. you have to do this only once.
  $sbe->generate_database_out_of_fastafile('ATH1.cdna', 'AGI Transcripts (- introns, + UTRs)');
 
  my %local_dbs_description = $sbe->get_databases();
  my @local_dbs = sort keys %local_dbs_description;
 
  # take first available database in our test
  $sbe->settings->database($local_dbs[0]);
 
  my $seq = 'UGAACAGAAAGCUCAUGAGCC'; 
 
  # search for the reverse complement and allow 4 mismatches
  # display 5 bases upstream and downstream of the match
  $sbe->settings->query($seq);
  $sbe->settings->reverse_complement(1);
  $sbe->settings->mismatches(4);
  $sbe->settings->upstream("5");
  $sbe->settings->downstream("5");
  
  # hypa does not generate alignments, but EMBOSS will automatically some for
  # you. If things should be fast, turn this off	
  # $sbe->settings->no_alignments(1);
 
  # with the features hash, you can check if the backend supports
  # a special feature. Allow wobble pairs (GU counts 0 or 0.5)	
  if (defined($sbe->features->{GUMISMATCHES})) {
 	  $sbe->settings->gumismatches(0.5); 		
  } else {
     print "\nBackend does not support wobble pairs\n";
  }	
 
  # now try to search. we use bioperls Exceptions
  try {
     $sbe->search();
  } catch Bio::Root::SystemException with {
     my $E = shift;
     print STDERR 'Backend call failed: ' . 	
     $E->{'-text'} . ' (' .  $E->{'-line'} . ")\n";
     exit(1);	
  } otherwise {        
 	my $E = shift;
     	print STDERR "An unexpected exception occurred: \n$E";
     exit(1);	
 
  };
 
  # output all informations we have!
  foreach my $res (@{$sbe->results}) {
     print $res->sequence->id . "\n";
     print $res->mark_subject_uppercase() . "\n";
     print $res->alignment_string() . "\n\n";
  }
   
=head1 DESCRIPTION

B<Weigel::Search::Backends::Hypa> searches for a query in a Hypa suffix array. 

NOTE 1: Hypa can not calculate alignments. But because we have the exact position
of the match, the alignment calculation shouldn't be too slow (In agrep, we align query 
and sequence, here query and approximate match).

NOTE 2: -online is available. But it is not recommended.

=head1 METHODS

See L<Weigel::Search::Backends::BackendI> for other methods. 

=over 2

=item Weigel::Search::Backends::Hypa-E<gt>new()

This function constructs a Hypa Backend object

   my $sbe = Weigel::Search::Backends::Hypa->new();

=cut

sub new {
   my $self = shift;
   $self = $self->SUPER::new;
   my %all_features = $self->features;
   delete $all_features{NATIVE_ALIGNMENTS};
   delete $all_features{EVALUE};
   delete $all_features{PERCENT_IDENTITY};
   $self->features(%all_features);
   $self;
}

=item $sbe-E<gt>search()

This function searches for the query specified in the L<Weigel::Search::Container::SearchSettings> object C< $sbe-E<gt>settings >
 
=cut

sub search {
   my $self           = shift;
   my $s              = $self->settings;
   my $tmp_query_file = $s->tmppath . '/spatter_' . time . '.hypa';

   $self->_prepare_query();
   $self->_check_search_settings();

   $self->_make_hypa_query_file($tmp_query_file);

   my $online = '';
   $online = ' -online ' if ( $s->online_isset && $s->online );

   my $upstream = '';
   $upstream = ' -showleft ' . $s->upstream . ' ' if $s->upstream > 0;
   my $downstream = '';
   $downstream = ' -showright ' . $s->downstream . ' ' if $s->downstream > 0;
   
   my $command =
       $self->_cat_path_filename( $s->execpath, 'hypa' ) . ' -showformat -q '
     . $tmp_query_file . $online . $upstream . $downstream . ' -index '
     . $self->_cat_path_filename( $s->datapath, $s->database );

   #print $command . " \n ";
   my $output = `$command`;

   # delete temporary files
   unlink($tmp_query_file);

   $self->throw( -class => 'Bio::Root::SystemException', 
                 -text  => "Hypa: Maybe query not valid. Command was:\n\t$command"
     )
     if ($?);
   $self->_output_parser($output);
   $self->_filter_results();
}

sub _make_hypa_query_file {
   my $self = shift;
   my $filename = shift;
   
   my $editops = ""; my $cost = "";
   my $query = $self->settings->query;
   my $gu_mm = 1; 
   $gu_mm = $self->settings->gumismatches if $self->settings->gumismatches_isset;
   my $mismatches = 0;
   $mismatches = $self->settings->mismatches if $self->settings->mismatches_isset;
   my $deletions = 0;
   $deletions = $self->settings->deletions if $self->settings->deletions_isset;
   my $insertions = 0;
   $insertions = $self->settings->insertions if $self->settings->insertions_isset;
   
   if ($gu_mm != 0.5 && $gu_mm != 1 && $gu_mm != 0) {
      $self->warn("Valid values for GU mismatches are: 0, 0.5 and 1");
   }

   if ($self->settings->editdistance_isset) {
      $self->warn("Editdistance not supported in this backend. Specify mismatches, insertions and deletions");
   }
   my $gu1 = 'G->T';
   my $gu2 = 'T->G';
  
   $self->settings->reverse_complement(0) unless $self->settings->reverse_complement_isset;
   if ($self->settings->reverse_complement) {
      $gu1 = 'A->G';
      $gu2 = 'C->T';
   }

   # what shall we do with GU?
   if ($gu_mm == 1) {    # normal mode,  a mismatch
   	 $editops = "[$mismatches, $deletions, $insertions]";
   } elsif ($gu_mm == 0) {   # not a mismatch
		$cost = <<EOT;
// match 0 * 2
cost(T->T) := 0;
cost(A->A) := 0;
cost(C->C) := 0;
cost(G->G) := 0;
cost($gu1) := 0;
cost($gu2) := 0;

// insertions, deletions, mismatches count 1 * 2
cost(.->) := 2;
cost(->.) := 2;
cost(.->T) := 2;
cost(.->C) := 2;
cost(.->G) := 2;
cost(.->A) := 2;

EOT
;
		$editops = "[cost, " . $mismatches*2 . "," . $deletions * 2 . "," . $insertions*2 . "]";
   } else  {  
		$cost = <<EOT;
// match 0 * 2
cost(T->T) := 0;
cost(A->A) := 0;
cost(C->C) := 0;
cost(G->G) := 0;

// gu 1/2 * 2 
cost($gu1) := 1;
cost($gu2) := 1;

// insertions, deletions, mismatches count 1 * 2
cost(.->) := 2;
cost(->.) := 2;
cost(.->T) := 2;
cost(.->C) := 2;
cost(.->G) := 2;
cost(.->A) := 2;

EOT
;
		$editops = "[cost, " . $mismatches*2 . "," . $deletions * 2 . "," . $insertions*2 . "]";
   }	   
   
	$query = "($query)";  
   open(FILE, "> $filename") || $self->throw(-class => 'Bio::Root::FileOpenException',
                                             -text  => "Can't open $filename for writing",
                                             -value => $! );
   my $hypaquery = <<EOT;
   $cost
   query = $query$editops;

EOT
   
   print FILE $hypaquery;
   close(FILE);
}   

=item $sbe->get_databases()

Returns a hash with all available databases. The keys are the filenames,
the values are descriptions (or the filename if no description is available).

Descriptions can be set in info files. For example, if you indexed file ATH1.cdna, Hypa constructs a lot of ATH1.cdna.* files. Now simply create a file 
ATH1.cdna.nfo and write a description in that file. The function C<generate_database_out_of_fastafile> will
create this file for you if you add a description as second argument.
 
=cut

sub get_databases {
   my $self = shift;
   return $self->_get_databases('.rev.suf');
}

=item $sbe->generate_database_out_of_fastafile('/tmp/ATH1.cdna')

Copies the specified file in the datapath directory (C<$sbe-E<gt>settings-E<gt>datapath>)
and generates a Hypa suffix array. You can get the databases with C<$sbe-E<gt>get_databases()>.

=cut

sub generate_database_out_of_fastafile {
   my $self        = shift;
   my $file        = shift;
   my $description = shift;
   my ( $filename, $oldpath ) = fileparse($file);
   
   my $olddir = getcwd();
   
   $self->_copy_fasta_file_and_create_nfo($file, $filename, $description);

   my $command =
     $self->_cat_path_filename( $self->settings->execpath, 'mkaffix.sh' )  
     . ' ' . $filename . ' ' . $filename
     . ' -dna';
   #print $command;
   system($command);
   $self->throw( -class => 'Bio::Root::SystemException', 
                 -text  => "Hypa error: Cannot generate suffix array. Command was:\n\t$command"
     )
     if ($?);
   # create database from directory of fasta files
   my $idx = Bio::Index::Fasta->new(
      -filename   => $filename . '.idx',
      -write_flag => 1
   );
   $idx->make_index( ($filename) );
   chdir($olddir);
}

sub _output_parser {
   my $self                = shift;
   my @lines               = split( "\n", shift );
   my $s                   = $self->settings;
   my @results             = ();
   
   # temp variables. for parsing only
   my $hit = -1;
   my $subject = '';
   my $match = '';
   my $tmp_seq;
   my $upstream_length = 0;

   foreach (@lines) {
       last if (/^Matches for pattern/);
       #print "LA" . $_ ."\n";
       if (/^\s*$/) {
            if ($hit == -1) {
               $hit = 1;
               next;
            }
            # first newline after hit, so now store hit.
            my $seq_query = Bio::Seq->new(
               -display_id => 'Query',
               -seq        => $s->query
            );
            # align only match
            my $seq_subject = Bio::Seq->new(
               -display_id => $tmp_seq->id,
               -seq        => $match
            );
            my $alignment = undef;
            $alignment = $self->_get_alignment( $seq_query, $seq_subject )
              unless $s->no_alignments;
            $s->querylength(length($s->query)) unless $s->querylength_isset;
            my $result =
              Weigel::Search::Container::SearchResult->new( $tmp_seq, $upstream_length,
               $upstream_length + $s->querylength,
               $alignment, $tmp_seq->id, '' );
            push( @results, $result );
            $self->results(@results);
            $hit = 1;
            next;
       }
       next if $hit == -1;  
       if ($hit == 1) {
         /^>(.*)/;
         $subject = $1;
         $hit = 0;
         next;
       }
       if (/hits$/) {
         next;
       }
       if (/^>(.*)\:\[(\d+)\,(\d+)\]/) {
            $tmp_seq = Bio::LocatableSeq->new(
               -id    => $1,
               -seq   => $subject,   # this is just a dummy, we will overwrite this in the next line
               -start => $2,
               -end   => $3
            );
         $subject = '';
         next;
      }
      $match = $_ if /^[^\^\$]/;
      if (/^\^/) {
         s/^\^//g;
         # maybe upstream region shorter than required
         $upstream_length = length($_);
      }
      s/^\$//g;
      $subject .= $_;
      $tmp_seq->seq($subject);

   }
}

=item $sbe-E<gt>get_sequences($sequence_ids) 

This function returns all sequences with the ids in the specified array reference as
a L<Bio::SeqIO> object. 


   my $seqio = $sbe->get_sequences([$id]);
   my $string;  my $stringio = IO::String->new($string);
   my $out = Bio::SeqIO->new('-fh' => $stringio,
                             '-format' => 'fasta');

   while ( my $seq = $seqio->next_seq() ) {
      # write the sequences in a string
      $out->write_seq($seq);
   }
   print $string;
 

=cut

sub get_sequences {
   my $self  = shift;
   my $seqid = shift;
   my $indexfile = $self->settings->datapath . '/' . $self->settings->database . '.idx';
   my $idx = Bio::Index::Fasta->new($indexfile);
   my $out = Bio::SeqIO->new( -format => 'Fasta' );
   foreach my $seqid ( @{$seqid} ) {
      $out->write_seq($idx->fetch($seqid));
   }
   return $out;
}

=back


=head1 SEE ALSO

L<Weigel::Search::Backends::BackendI>
L<Weigel::Search::Container::SearchSettings>
L<Bio::SeqIO>

=head1 AUTHOR

Markus Riester, <markus.riester@tuebingen.mpg.de@>

=cut

1;

# vim: ft=perl sw=3 ts=3 expandtab
