package Weigel::Search::Backends::Vmatch;

use strict;
use warnings;

use Weigel::Search::Container::SearchSettings;
use Weigel::Search::Container::SearchResult;
use Weigel::Search::Backends::BackendI;

use base 'Weigel::Search::Backends::BackendI';

use Bio::Seq;
use Bio::SeqIO;

use File::Basename;
use IO::String;
use File::Copy;
use Cwd;

our $VERSION = '0.3';

=head1 NAME

Backends::Vmatch.pm - Vmatch Backend  

=head1 SYNOPSIS

  use Weigel::Search::Backends::Vmatch;
  
  use Bio::Root::Exception;
  use Error qw(:try);
  
  # configure our search backend, in this case Vmatch
  my $sbe = Weigel::Search::Backends::Vmatch->new();
  
  $sbe->settings->execpath('/usr/local/vmatch');
  $sbe->settings->tmppath('/tmp');
  $sbe->settings->datapath('data');
  
  # generate a Vmatch suffix array. you have to do this only once.
  $sbe->generate_database_out_of_fastafile('ATH1.cdna', 'AGI Transcripts (- introns, + UTRs)');
  
  my %local_dbs_description = $sbe->get_databases();
  my @local_dbs = sort keys %local_dbs_description;
  
  # take first available database in our test
  $sbe->settings->database($local_dbs[0]);
  
  # search for the reverse complement and allow 4 mismatches
  $sbe->settings->query('UGAACAGAAAGCUCAUGAGCC');
  $sbe->settings->reverse_complement(1);
  $sbe->settings->mismatches(4);
  
  # With many mismatches and short queries, the "online" algorithm
  # is maybe faster. This alogrithm does not use the index. Test this!
  $sbe->settings->online(1);
  
 # now try to search. we use bioperls Exceptions
 try {
    $sbe->search();
 } catch Bio::Root::SystemException with {
    my $E = shift;
    print STDERR 'Backend call failed: ' . 	
    $E->{'-text'} . ' (' .  $E->{'-line'} . ")\n";
    exit(1);	
 } otherwise {        
	my $E = shift;
    	print STDERR "An unexpected exception occurred: \n$E";
      exit(1);	
 };

  # output the searchresults with nice alignments
  foreach my $res (@{$sbe->results}) {
     print $res->sequence->id . "\n";
     print $res->mark_subject_uppercase() . "\n";
     print $res->alignment_string() . "\n\n";
  }
  
=head1 DESCRIPTION

B<Weigel::Search::Backends::Vmatch> searches for a query in a VMATCH suffix array. 

=head1 METHODS

See L<Weigel::Search::Backends::BackendI> for other methods. 

=over 2

=item Weigel::Search::Backends::Vmatch-E<gt>new()

This function constructs a Vmatch Backend object

   my $sbe = Weigel::Search::Backends::Vmatch->new();

=cut

sub new {
   my $self = shift;
   $self = $self->SUPER::new;
   my %all_features = $self->features;
   delete $all_features{GUMISMATCHES};
   delete $all_features{DELETIONS};
   delete $all_features{INSERTIONS};
   $self->features(%all_features);
   $self;
}

=item $sbe-E<gt>search()

This function searches for the query specified in the L<Weigel::Search::Container::SearchSettings> object C< $sbe-E<gt>settings >
 
=cut

sub search {
   my $self           = shift;
   my $s              = $self->settings;
   my $tmp_query_file = $s->tmppath . '/spatter_' . time;

   $self->warn("GU mismatch options not available in this backend")
     if $s->gumismatches_isset && $s->gumismatches != 1;
   
   $self->_prepare_query();
   $self->_check_search_settings();

   # construct a temporary fasta file with the query for vmatch
   my $seqobj = Bio::Seq->new(
      -display_id => $s->query,
      -seq        => $s->query
   );

   my $outseqio = Bio::SeqIO->new(
      -file   => ">$tmp_query_file",
      -format => 'fasta'
   );

   $outseqio->write_seq($seqobj);

   # now generate the command string

   my $fuzzy = "";
   $fuzzy = "-h " . $s->mismatches if $s->mismatches > 0;
   if ( $s->editdistance_isset ) {
      $fuzzy = "-e " . $s->editdistance;
      $self->warn("Ignoring mismatch settings because edit distance is set.")
        if $s->mismatches > 0;
   }
   my $online = '';
   $online = ' -online ' if ( $s->online_isset && $s->online );
   $s->querylength( length( $s->query ) ) unless $s->querylength_isset;
   my $length = "-l " . $s->querylength;

   my $command =
       $self->_cat_path_filename( $s->execpath, 'vmatch' ) . ' -q '
     . $tmp_query_file
     . " $fuzzy -best "
     . $s->maxhits . ' '
     . $length . ' -s '
     . $online
     . $self->_cat_path_filename( $s->datapath, $s->database );

   print $command . " \n ";
   my $output = `$command`;

   # delete temporary files
   unlink($tmp_query_file);

   $self->throw( -class => 'Bio::Root::SystemException', 
                 -text  => "Vmatch error: Query not valid(maybe too many mismatches? Try online(1))\n\nCommand was:\n\t$command"
     )
     if ($?);
   $self->_output_parser($output);
   $self->_filter_results();
}

=item $sbe->get_databases()

Returns a hash with all available databases. The keys are the filenames,
the values are descriptions (or the filename if no description is available).

Descriptions can be set in info files. For example, if you indexed file ATH1.cdna, vmatch constructs a lot of ATH1.cdna.* files. Now simply create a file 
ATH1.cdna.nfo and write a description in that file. The function C<generate_database_out_of_fastafile> will
create this file for you if you add a description as second argument.
 
=cut

sub get_databases {
   my $self = shift;
   return $self->_get_databases('.al1');
}

=item $sbe->generate_database_out_of_fastafile('/tmp/ATH1.cdna')

Copies the specified file in the datapath directory (C<$sbe-E<gt>settings-E<gt>datapath>)
and generates a VMATCH suffix array. You can get the databases with C<$sbe-E<gt>get_databases()>.

=cut

sub generate_database_out_of_fastafile {
   my $self        = shift;
   my $file        = shift;
   my $description = shift;
   my ( $filename, $oldpath ) = fileparse($file);

   my $olddir = getcwd();
   
   $self->_copy_fasta_file_and_create_nfo($file, $filename, $description);

   my $command =
     $self->_cat_path_filename( $self->settings->execpath, 'mkvtree' ) . ' -db '
     . $filename
     . " -dna -pl 3 -allout -v ";
   print $command;
   system($command);
   $self->throw( -class => 'Bio::Root::SystemException', 
                 -text  => "Vmatch error: Cannot generate suffix array. Command was:\n\t$command"
     )
     if ($?);
   chdir($olddir);
}

sub _output_parser {
   my $self                = shift;
   my @lines               = split( "\n", shift );
   my $s                   = $self->settings;
   my @results             = ();
   my $alignment_in_output = -1;
   my $cnt                 = 0;
   my $skip_next_alignment = 0;
   my $subject;

   foreach (@lines) {
      s/\s+/ /g;
      next unless /\s/;
      my @fields = split(" ");
      $alignment_in_output = 0 if ( /^Sbjct:/ && !$skip_next_alignment );

      next unless ( $fields[0] =~ /^\d+$/ || $alignment_in_output >= 0 );
      $skip_next_alignment = 0;

      if (/^Sbjct: (.*)$/) {
         $subject = $1;
      }
      if (/^Query: (.*)$/) {
         my $query = $1;
         $query =~ s/\s+(\d+)$//;
         my $query_pos = $1;
         $subject =~ s/\s+(\d+)$//;
         my $subject_pos = $1;
         my $tmp_aln = new Bio::SimpleAlign( -source => "VMATCH" );
         $tmp_aln->add_seq(
            Bio::LocatableSeq->new(
               -id    => " Subject ",
               -seq   => $subject,
               -start => ( $subject_pos - length($subject) ) + 1,
               -end   => $subject_pos
            )
         );
         $tmp_aln->add_seq(
            Bio::LocatableSeq->new(
               -id    => "Query",
               -seq   => $query,
               -start => ( $query_pos - length($query) ) + 1,
               -end   => $query_pos
            )
         );
         $results[ $cnt - 1 ]->alignment($tmp_aln);
         $alignment_in_output = -1;
         next;
      }
      next if $alignment_in_output >= 0;

      my $start = $fields[2] - $s->upstream;

      # maybe the defined upstream region is larger than available
      # so check this and store in local variables
      my $upstream = $s->upstream;
      if ( $start < 0 ) {
         $upstream = $upstream + $start;
         $start    = 0;
      }
      my $length  = $upstream + $fields[0] + $s->downstream;
      my $command =
          $self->_cat_path_filename( $s->execpath, 'vsubseqselect' )
        . " -seq $length $fields[1] "
        . $start . ' '
        . $self->_cat_path_filename( $s->datapath, $s->database );
      my $output = `$command`;
      my $seq    = $output;
      $seq =~ /^(.*?)\n/;
      my $desc = $1;
      $desc =~ s/^>//;

      $seq =~ s/>.*//g;
      $seq =~ s/^.*?\n//;
      $seq =~ s/\n//g;

      my $fasta = Bio::Seq->new(
         -display_id => $desc,
         -seq        => $seq
      );

      my $query = substr $seq, $upstream, $fields[0];

      my $seq_query = Bio::Seq->new(
         -display_id => " Query ",
         -seq        => $s->query
      );

      my $seq_subject = Bio::Seq->new(
         -display_id => " Subject ",
         -seq        => $query
      );

      my $alignment =
        Bio::SimpleAlign->new()
        ;    #$self->_get_alignment($seq_query, $seq_subject);
      my $result =
        Weigel::Search::Container::SearchResult->new( $fasta, $s->upstream,
         $s->upstream + $s->querylength,
         $alignment, $fields[1], '' );
      $result->evalue($fields[8]);
      $result->percent_identity($fields[10]);
      $cnt++;
      push( @results, $result );
      $self->results(@results);
   }
}

=item $sbe-E<gt>get_sequences($sequence_ids) 

This function returns all sequences with the ids in the specified array reference as
a L<Bio::SeqIO> object. 


   my $seqio = $sbe->get_sequences([$id]);
   my $string;  my $stringio = IO::String->new($string);
   my $out = Bio::SeqIO->new('-fh' => $stringio,
                             '-format' => 'fasta');

   while ( my $seq = $seqio->next_seq() ) {
      # write the sequences in a string
      $out->write_seq($seq);
   }
   print $string;
 

=cut

sub get_sequences {
   my $self    = shift;
   my $seqid   = shift;
   my $s       = $self->settings;
   my $tmpfile = $self->_cat_path_filename($s->tmppath , 'vseqselect' . time);

   open( FILE, " > $tmpfile " );
   foreach ( @{$seqid} ) {
      print FILE $_ . " \n ";
   }
   close(FILE);
   my $command =
       $self->_cat_path_filename( $s->execpath, 'vseqselect' )
     . " -seqnum $tmpfile "
     . $self->_cat_path_filename( $s->datapath, $s->database );

   #   print $command;
   my $output = `$command`;
   $self->throw( -class => 'Bio::Root::SystemException', 
                 -text  => "Vmatch error: Cannot fetch sequence out of suffix array. Command was:\n\t$command"
     )
     if ($?);
   unlink($tmpfile);
   my $stringio = IO::String->new($output);
   my $out      = Bio::SeqIO->new(
      '-fh'     => $stringio,
      '-format' => 'fasta'
   );
   return $out;
}

=back


=head1 SEE ALSO

L<Weigel::Search::Backends::BackendI>
L<Weigel::Search::Container::SearchSettings>
L<Bio::SeqIO>

=head1 AUTHOR

Markus Riester, <markus.riester@tuebingen.mpg.de@>

=cut

1;

# vim: ft=perl sw=3 ts=3 expandtab
