package Weigel::Search::Container::SearchResult;

use strict;
use warnings;

use IO::String;
use Carp qw(carp);

use base 'Bio::Root::RootI';

our $VERSION = '0.1';

use Class::MethodMaker [
   new    => 'new2',
   scalar => [qw / sequence begin end alignment sequence_id remark percent_identity evalue dG/],
];

=head1 NAME

Weigel::Search::Container::SearchResult - data structure for a backend search hit

=head1 SYNOPSIS
  
See Backends::Backend::Vmatch for example code 

=head1 DESCRIPTION

This is the datastructure for one hit in the database.

=head1 METHODS

=over 2

=item new(candidate, alignment, begin, end, alignment, sequence_id, remark );

This function constructs a SearchResult object. 

=cut

sub new {
   my $self = shift->new2;

   #initizialize member variables

   #from params
   $self->sequence(shift);
   $self->begin(shift);
   $self->end(shift);
   $self->alignment(shift);
   $self->sequence_id(shift);
   $self->remark(shift);
   $self;
}

=item mark_subject_uppercase()

This function returns the sequence in a string. the substring from $self->begin to
$self->end will be in uppercase, the rest in lowercase

=cut

sub mark_subject_uppercase {
   my $self   = shift;
   my $result = $self->sequence->seq;
   return
     lc( substr( $result, 0, $self->begin ) )
     . uc( substr( $result, $self->begin, $self->end - $self->begin ) )
     . lc( substr( $result, $self->end ) );
}

=item subject()

Get the sequence found in database as string without upstream and downstream regions

=cut

sub subject {
   my $self = shift;

   return Bio::Seq->new(
      -seq => $self->sequence->subseq( $self->begin + 1, $self->end ) );
}

=item alignment_string()

This function returns a string with the formated alignment. We use CLUSTALW
Format without many blank lines and CLUSTAL header. In some backends like
Agrep, this function will return an empty string if no_alignments is true
in the backend search settings (L<Weigel::Search::Container::SearchSettings>).

=cut

sub alignment_string {
   my $self   = shift;
   my $result = "";
   my $str    = IO::String->new();
   my $out    = Bio::AlignIO->new( -format => 'clustalw', -fh => $str );
   unless (defined ($self->alignment)) {
      $self->warn("No alignment calculated.");
      return "";
   }   
   $out->write_aln( $self->alignment );
   $out->close();
   $str = IO::String->new( ${ $str->string_ref } );
   while ( my $l = <$str> ) {
      $result .= $l unless ( $l =~ /CLUSTAL/ or $l =~ /^\s+$/ );
   }
   return $result;
}

=item alignment_to_csv()

This function generates a string for a CSV file, where every base of the stored alignment gets its own column. Additional infos
like description and dG are displayed, too.

=cut

sub alignment_to_csv {

   # gets an alignment and returns and string for a CSV-file
   my $self = shift;

   my $aln = $self->alignment;
   unless (defined ($aln)) {
      $self->warn("No alignment calculated.");
      return "";
   }   
   return $self->get_gene_id() . ','
     . $_->_commafy_string( $aln->get_seq_by_pos(1)->seq ) . ","
     . $self->sequence->id . "\n,"
     . $self->_commafy_string( $aln->match_line )
     . ",\nQuery" . ","
     . $_->_commafy_string( $aln->get_seq_by_pos(2)->seq )
     . ",dG = "
     . $self->dG . "\n";

}

=item get_gene_id()

This function tries to guess gene name out of sequencename

=cut

sub get_gene_id {
   my $self = shift;
   my $desc = $self->sequence->id;
   $desc =~ s/^(.*?)( |\|)//;
   return $1;
}

sub _commafy_string {

   # param: string
   # output: s,t,r,i,n,g
   my $self = shift;
   my $s    = shift;

   my $result = "";
   for ( my $i = 0 ; $i != length($s) ; ++$i ) {
      $result .= substr( $s, $i, 1 ) . ',';
   }
   chop($result);
   return $result;
}

=item sequence()

Get/set the sequence found in database. L<Bio::Seq> object.

=item sequence_id()

Getter/Setter for sequence ID in database. This is an internal ID of the
backend, not any ID of some annotation in the sequence name. Use
get_gene_id() to guess this ID. The internal ID can be used in the
backend function get_sequences (L<Weigel::Search::Backends::Backend>).

=item begin()

Get/set the position of the beginning of the subject in the sequence. This allows 
retrieving upstream regions from the backend.

=item end()

Get/set the position of the end of the subject in the sequence. This allows retrieving
downstream regions from the backend.

=item alignment()

Get/set the alignment of the match. See L<Bio::SimpleAlign> for details. There are powerful
modules for converting this modul in many formats. See L<Bio::AlignIO> for details.

=item dG()

Get/set  dG . See L<Weigel::Search::RNA::Hybridization> for details.

=item remark()

Get/set some additional informations like filter results to this hit.

=back

=head1 SEE ALSO

L<Bio::SimpleAlign> 
L<Bio::AlignIO> 
L<Bio::Seq>
L<Bio::SeqIO>
L<Weigel::Search::RNA::Hybridization>
L<Weigel::Search::Backend::Backends>

=head1 AUTHOR

Markus Riester, <markus.riester@tuebingen.mpg.de>

=cut

1;

# vim: ft=perl sw=3 ts=3 expandtab
