package Weigel::Search::Container::SearchSettings;

use strict;
use warnings;

our $VERSION = '0.2';

use Class::MethodMaker [
   new    => 'new2',
   scalar => [
      qw / mismatches insertions deletions editdistance query querylength subject gumismatches upstream downstream maxhits filter no_alignments datapath database online tmppath execpath reverse_complement /
   ],
];

=head1 NAME

Weigel::Search::Container::SearchSettings.pm - an object for storing all search settings 

=head1 SYNOPSIS

 
 use Weigel::Search::Backends::Vmatch;
 use Weigel::Search::Filter::MIRNAFilter;

 # configure our search backend, in this case VMATCH
 my $sbe = Weigel::Search::Backends::Vmatch->new();
 $sbe->path("/usr/local/virtual.distrib");

 # the backend automatically generates a SearchSettings
 # object. You can use this now to configure the
 # backend:  
 $sbe->settings->tmppath($tmp_path);
 my $filter = Weigel::Search::Filter::MIRNAFilter->new();
 $sbe->settings->filter($filter);
 $sbe->settings->mismatches(4);
 $sbe->settings->datapath($path_to_data_directory);
 $sbe->settings->database("ATH1.cdna");
 
 # generate a Bioperl sequence object
 my $seq = Bio::Seq->new( -display_id => "Seq A",
        		    -seq => "UGAACAGAAAGCUCAUGAGCC");
 
 # search for the reverse complement 
 $sbe->settings->query($seq->revcom->seq);

=head1 DESCRIPTION

Backends::SearchSettings is a datastructure for storing all search settings.
Not all backends will support every option.

=head1 METHODS

=over 2

=item new()

This function constructs a SearchSettings object

=cut

sub new {
   my $self = shift->new2;

   # initialize standard settings
   $self->mismatches(0);
   $self->insertions(0);
   $self->deletions(0);
   $self->gumismatches(1);
   $self->no_alignments(0);
   $self->maxhits(75000);
   $self->upstream(0);
   $self->downstream(0);
   $self->reverse_complement(0);
   $self;
}

=item to_string() 

This function returns a string representation of this object

=cut

sub to_string {
   my $self   = shift;
   my $result = "";
   $result .= "Mismatches   : " . $self->mismatches . "\n"
     if $self->mismatches_isset;
   $result .= "Insertions   : " . $self->insertions . "\n"
     if $self->insertions_isset;
   $result .= "Deletions    : " . $self->deletions . "\n"
     if $self->deletions_isset;
   $result .= "Edit distance: " . $self->editdistance . "\n"
     if $self->editdistance_isset;
   $result .= "No alignments: " . $self->no_alignments . "\n"
     if $self->no_alignments_isset;
   $result .= "Query        : " . $self->query . "\n" if $self->query_isset;
   $result .= "Querylength  : " . $self->querylength . "\n"
     if $self->querylength_isset;
   $result .= "Subject      : " . $self->subject . "\n" if $self->subject_isset;
   $result .= "GU Mismatches: " . $self->gumismatches . "\n";
   $result .= "upstream     : " . $self->upstream . "\n";
   $result .= "downstream   : " . $self->downstream . "\n";
   $result .= "maxhits      : " . $self->maxhits . "\n";
   $result .= "filter       : " . $self->filter . "\n" if $self->filter_isset;
   $result .= "Data Path    : " . $self->datapath . "\n";
   $result .= "Database     : " . $self->database . "\n";
   $result .= "Tmp Path     : " . $self->tmppath . "\n";
   return $result;
}

=item mismatches()

Get/Set allowed mismatches

=item editdistance()

Get/Set allowed editdistance. Not all backends will support this feature!

=item query()

Get/Set the query, a simple string. Maybe this will change in future
versions to allow multiple queries

=item querylength()

Get/Set the query length. Initialized with length of the query string. If
this member variable is smaller then the query string, then the backend will
search for all substrings of that size. Not all backends will support this
feature!

=item upstream()

Get/set upstream. This is the number of bases upstream the match.
 
=item downstream()

Get/set downstream. This is the number of bases downstream the match.

=item filter()

Get/set the filter. This is a module based on Weigel::Search::Filter::Filter.

=item datapath()

Get/set the datapath. This is the path to the databases for the backend.

=item database()

Get/set the database. This is the name of the database in $self->datapath

=item tmppath()

Get/set the tmppath. This is a path were the backend can store temporary files.

=item execpath()

Get/set the execpath. This is a path to the backend executable.

=item no_alignments()

Get/set the no_alignments. Some backends like Agrep don't output alignments. 
The EMBOSS Smith-Waterman implementation will automatically generate alignments
for a search result. For perfomance reasons, you can turn turn this feature off:

  # if things need to be fast, turn alignments off
  $sbe->settings->no_alignments(1);

Not available in the Vmatch Backend.

=item online()

Get/set online. Vmatch backend allows searching without using the index. When
you allow many mismatches, this could be faster. 

=back

=head1 SEE ALSO

L<Weigel::Search::Filter::Filter>
L<Weigel::Search::Backends::Backend>

=head1 AUTHOR

Markus Riester, <markus.riester@tuebingen.mpg.de@>

=cut

1;

# vim: ft=perl sw=3 ts=3 expandtab
